"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Dict, Any
import numpy as np
import cv2


def draw_outputs(img: List, outputs: List[Any], class_names: Dict[str, str]) -> List:
    """Draw object bounding box, confident score, and class name on
    the object in the image.

    Input:
        - img:      the image
        - outputs:  the outputs of prediction, which contain the
                    object' bounding box, confident score, and class
                    index
        - class_names: dictionary of classes string names

    Output:
        - img:      the image with the draw of outputs
    """
    boxes, objectness, classes = outputs
    width_height = np.flip(img.shape[0:2])
    for i, oneclass in enumerate(classes):
        x1y1 = tuple((np.array(boxes[i][0:2]) * width_height).astype(np.int32))
        x2y2 = tuple((np.array(boxes[i][2:4]) * width_height).astype(np.int32))
        img = cv2.rectangle(img, x1y1, x2y2, (255, 0, 0), 2)
        img = cv2.putText(img, '{} {:.4f}'.format(class_names[oneclass], objectness[i]),
                          x1y1, cv2.FONT_HERSHEY_COMPLEX_SMALL, 1, (0, 0, 255), 2)
    return img



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import zipfile
import requests
from tqdm import tqdm

def download_weights(root, url):
    """Download weights for specified url

    Args:
        root (str): root directory of peekingduck
        url (str): url to download weights from
    """

    extract_dir = os.path.join(root, "..", "weights")
    zip_path = os.path.join(root, "..", "weights", "temp.zip")

    download_file_from_google_drive(url, zip_path)

    # search for downloaded .zip file and extract, then delete
    with zipfile.ZipFile(zip_path, "r") as temp:
        for file in tqdm(iterable=temp.namelist(), total=len(temp.namelist())):
            temp.extract(member=file, path=extract_dir)

    os.remove(zip_path)

def download_file_from_google_drive(file_id, destination):
    """Method to download publicly shared files from google drive

    Args:
        file_id (str): file id of google drive file
        destination (str): destination directory of download
    """
    url = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(url, params={'id': file_id}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id': file_id, 'confirm': token}
        response = session.get(url, params=params, stream=True)

    save_response_content(response, destination)

def get_confirm_token(response):
    """Method to get confirmation token

    Args:
        response (Response): html response from call

    Returns:
        value (str): confirmation token
    """
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def save_response_content(response, destination):
    """Chunk saving of download content. Chunk size set to large
    integer as weights are usually pretty large

    Args:
        response (Reponse): html response
        destination (str): destintation directory of download
    """
    chunk_size = 32768

    with open(destination, "wb") as temp:
        for chunk in tqdm(response.iter_content(chunk_size)):
            if chunk: # filter out keep-alive new chunks
                temp.write(chunk)



"""Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os
from typing import Any, Dict
import numpy as np
import cv2
from peekingduck.pipeline.nodes.node import AbstractNode

# role of this node is to be able to take in multiple frames, stitch them together and output them.
# to do: need to have 'live' kind of data when there is no filename
# to do: it will be good to have the accepted file format as a configuration
# to do: somewhere so that input and output can use this config for media related issues


class Node(AbstractNode):
    """Node that processes videos and images as primary source input"""
    def __init__(self, config):
        super().__init__(config, node_path=__name__)

        self._file_name = None
        self._output_dir = config["outputdir"]
        self._prepare_directory(config["outputdir"])
        self._fourcc = cv2.VideoWriter_fourcc(*'mp4v')

        #initialize for use in run
        self._file_name = None
        self._file_path = None
        self._image_type = None
        self.writer = None

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ Writes media information to filepath

        Args:
            inputs: ["filename", "img", "fps"]

        Returns:
            outputs: [None]
        """

        if not self._file_name:
            self._prepare_writer(inputs["filename"],
                                 inputs["img"],
                                 inputs["fps"])

        if inputs["filename"] != self._file_name:
            self._prepare_writer(inputs["filename"],
                                 inputs["img"],
                                 inputs["fps"])
        else:
            self._write(inputs["img"])

        return {}

    def _write(self, img: np.array):
        if self._image_type == "image":
            cv2.imwrite(self._file_path, img)
        else:
            self.writer.write(img)

    def _prepare_writer(self, filename: str, img: np.array, fps: int):

        self._file_name = filename
        self._image_type = "video"
        if filename.split(".")[-1] in ["jpg", "jpeg", "png"]:
            self._image_type = "image"

        resolution = img.shape[1], img.shape[0]
        self._file_path = os.path.join(self._output_dir, filename)
        self.writer = cv2.VideoWriter(self._file_path, self._fourcc, fps, resolution)

    @staticmethod
    def _prepare_directory(outputdir):
        os.makedirs(outputdir, exist_ok=True)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """This is a template class of how to write a node for peekingduck."""
    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__(config, node_path=__name__)

        # initialize/load any configs and models here
        # self.config1 = config["config1"]
        # self.logger.info("model loaded with configs: %s", configs)


    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """ This node does ___.

        Args:
            inputs (dict): Dict with keys "__", "__".

        Returns:
            outputs (dict): Dict with keys "__".
        """

        # result = do_something(inputs["in1"], inputs["in2"])
        # outputs = {"out1": result}
        # return outputs



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from threading import Thread, Lock
import cv2
from peekingduck.pipeline.nodes.input.utils.preprocess import set_res, mirror


class VideoThread:
    '''
    Videos will be threaded to prevent I/O blocking from affecting FPS.
    '''

    def __init__(self, res, input_source, mirror_image):
        self.stream = cv2.VideoCapture(input_source)
        self.mirror = mirror_image
        if not self.stream.isOpened():
            raise ValueError("Camera or video input not detected: %s" % input_source)

        width, height = res['width'], res['height']
        set_res(self.stream, width, height)
        self._lock = Lock()

        self.frame = None
        thread = Thread(target=self._reading_thread, args=(), daemon=True)
        thread.start()

    def __del__(self):
        self.stream.release()

    def _reading_thread(self):
        '''
        A thread that continuously polls the camera for frames.
        '''
        while True:
            _, self.frame = self.stream.read()

    def read_frame(self):
        '''
        Reads the frame.
        '''
        self._lock.acquire()
        if self.frame is not None:
            frame = self.frame.copy()
            self._lock.release()
            if self.mirror:
                frame = mirror(frame)
            return True, frame

        self._lock.release()
        return False, None


class VideoNoThread:
    '''
    No threading to deal with recorded videos and images.
    '''

    def __init__(self, res, input_source, mirror_image):
        self.stream = cv2.VideoCapture(input_source)
        self.mirror = mirror_image
        if not self.stream.isOpened():
            raise ValueError("Video or image path incorrect: %s" % input_source)

        width, height = res['width'], res['height']
        set_res(self.stream, width, height)

    def __del__(self):
        self.stream.release()

    def read_frame(self):
        '''
        Reads the frame.
        '''
        return self.stream.read()

    @property
    def fps(self):
        """ Get FPS of videofile

        Returns:
            int: number indicating FPS
        """
        fps = self.stream.get(cv2.CAP_PROP_FPS)
        return fps



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from typing import Any, Dict, List
import yaml
from peekingduck.pipeline.nodes.node import AbstractNode

class ConfigLoader:
    """ Reads configuration and returns configuration required to
    create the nodes for the project
    """
    def __init__(self, nodes: List[AbstractNode], node_yml: str = None):
        self._master_node_config = {}
        self._rootdir = os.path.join(
            os.path.dirname(os.path.abspath(__file__))
        )
        if node_yml:
            node_config = self._load_from_path(node_yml)
            self._master_node_config = node_config
        else:
            self.load_from_node_list(nodes)

    def load_from_node_list(self, nodes: List[AbstractNode]) -> None:
        """load node_configs from a list of nodes.
        Configs is expected to be at level of peekingduck"""
        for node in nodes:
            node_type, node_name = node.split('.')
            config_filename = node_name + '.yml'
            filepath = os.path.join(self._rootdir, 'configs', node_type, config_filename)
            node_config = self._load_from_path(filepath)
            self._master_node_config[node] = node_config

    @staticmethod
    def _load_from_path(filepath: str) -> None:
        """load node_configs directly from a custom node_config"""
        with open(filepath) as file:
            node_config = yaml.load(file, Loader=yaml.FullLoader)
        return node_config

    def get(self, item: str) -> Dict[str, Any]:
        """Get item from configuration read from the filepath,
        item refers to the node item configuration you are trying to get"""

        node = self._master_node_config[item]

        # some models require the knowledge of where the root is for loading
        node['root'] = self._rootdir
        return node



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List, Set, Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode


class Pipeline:
    """ Pipe class that stores nodes and manages the data information used during inference
    """

    def __init__(self, nodes: List[AbstractNode]):
        """
        Args:
            nodes (:obj:'list' of :obj:'Node'): node stack as declared for use in
                inference pipeline
        """
        self._check_pipe(nodes)
        self.nodes = nodes
        self._datas = {}
        self.video_end = False

    def execute(self) -> None:
        """ executes all node contained within the pipe
        """

        for node in self.nodes:
            inputs = {key: self._datas[key]
                      for key in node.inputs if key in self._datas}
            outputs = node.run(inputs)

            if 'end' in outputs and outputs['end']:
                self.video_end = True
                break

            self._datas.update(outputs)

    def get_pipeline_results(self) -> Dict[str, Any]:
        """get all results data of nodes in pipeline

        Returns:
            Dict[Any]: Dictionary of all pipeline node results
        """
        return self._datas

    @staticmethod
    def _check_pipe(nodes: Set[AbstractNode]):
        # 1. Check the initial node is a source node
        # 2. Check every subsequent node utilizes something that will exist
        # if reached end, it is all valid

        data_pool = []

        if nodes[0].inputs[0] == 'source':
            data_pool.extend(nodes[0].outputs)

        for node in nodes[1:]:

            if all(item in data_pool for item in node.inputs):
                data_pool.extend(node.outputs)
            else:
                raise ValueError(
                    'Graph nodes do not form proper channel. Please check nodes.'
                )



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict
from time import perf_counter

from peekingduck.pipeline.nodes.node import AbstractNode
from .utils.drawfunctions import draw_fps

class Node(AbstractNode):
    """ FPS node class that calculates the FPS and draw the FPS onto the image
    frame
    """
    def __init__(self, config: Dict) -> None:
        super().__init__(config, node_path=__name__)

        self.previous_time = 0

    def run(self, inputs: Dict[str, Any]) -> None:
        """ Calculates FPS using the time difference between the current frame
        and the previous frame. Calculated FPS is then draw onto image frame

        Args:
            inputs: ["img"]

        Returns:
            outputs: [None]
        """

        current_time = perf_counter()
        current_fps = 1 / (current_time - self.previous_time)
        self.previous_time = current_time
        draw_fps(inputs['img'], current_fps)

        return {}



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Any
import numpy as np
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Node that used 2d bounding boxes information to estimate 3d location"""
    def __init__(self, config: Dict) -> None:
        super().__init__(config, node_path=__name__)

        self.height_factor = config['height_factor']
        self.focal_length = config['focal_length']

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts 2D bounding boxes into 3D locations.

        Args:
            inputs (dict): Dict with keys "bboxes".

        Returns:
            outputs (dict): Dict with keys "obj_3D_locs".
        """

        locations = []

        for idx, bbox in enumerate(inputs["bboxes"]):
            # Subtraction is to make the camera the origin of the coordinate system
            center_2d = (bbox[0:2] + bbox[2:4] * 0.5) - np.array([0.5, 0.5])

            z_coord = (self.focal_length * self.height_factor) / bbox[3]
            x_coord = (center_2d[0] * self.height_factor) / bbox[3]
            y_coord = (center_2d[1] * self.height_factor) / bbox[3]

            point = np.array([x_coord, y_coord, z_coord])
            locations.append({"idx": idx, "3D_loc": point})

        outputs = {"obj_3D_locs": locations}

        return outputs



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Any

class AbstractNode(metaclass=ABCMeta):
    """Abstract Node class that defines requirements for a node."""

    def __init__(self, config: dict, node_path: str):

        self._inputs = config['input']
        self._outputs = config['output']
        self._name = node_path

        self.logger = logging.getLogger(self._name)

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'run') and
                callable(subclass.run))

    @abstractmethod
    def run(self, inputs: Dict[str, Any]):
        """abstract method needed for running node"""
        raise NotImplementedError("This method needs to be implemented")

    @property
    def inputs(self) -> List[str]:
        """getter for input requirements"""
        return self._inputs

    @property
    def outputs(self) -> List[str]:
        """getter for node outputs"""
        return self._outputs

    @property
    def name(self) -> str:
        """getter for node name"""
        return self._name



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
import yaml
import click
from peekingduck.runner import Runner
from .utils.logger import setup_logger

setup_logger()
logger = logging.getLogger(__name__) #pylint: disable=invalid-name

@click.group()
def cli():
    """
    PeekingDuck is a python framework for dealing with Machine Learning model inferences.

    Developed by Computer Vision Hub at AI Singapore.
    """

def _get_cwd():
    return os.getcwd()

def create_custom_folder():
    """Make custom nodes folder to create custom nodes
    """
    curdir = _get_cwd()
    custom_node_dir = os.path.join(curdir, "src/custom_nodes")
    logger.info("Creating custom nodes folder in %s", custom_node_dir)
    os.makedirs(custom_node_dir, exist_ok=True)


def create_yml():
    """Inits the declarative yaml"""
    #Default yml to be discussed
    default_yml = dict(
        nodes=[
            'input.live',
            'model.yolo',
            'draw.bbox',
            'output.screen'
        ]
    )

    with open('run_config.yml', 'w') as yml_file:
        yaml.dump(default_yml, yml_file, default_flow_style=False)

@cli.command()
def init():
    """Initialise a PeekingDuck project"""
    print("Welcome to PeekingDuck!")
    create_custom_folder()
    create_yml()


@cli.command()
@click.option('--config_path', default=None, type=click.Path(),
              help="List of nodes to run. None assumes \
                   run_config.yml at current working directory")
def run(config_path: str) -> None:
    """Runs PeekingDuck"""
    curdir = _get_cwd()
    if not config_path:
        config_path = os.path.join(curdir, "run_config.yml")

    custom_node_path = os.path.join(curdir, 'src/custom_nodes')

    runner = Runner(config_path, custom_node_path)
    runner.run()


@cli.command()
@click.option('--config_path', default=None, type=click.Path(),
              help="List of nodes to pull config ymls from. \
                   If none, assumes a run_config.yml at current working directory")
def get_configs(config_path: str) -> None:
    """Creates node specific config ymls for usage. If no configs are specified, pull all"""
    if not config_path:
        curdir = _get_cwd()
        config_path = os.path.join(curdir, "run_config.yml")

    with open(config_path) as node_yml:
        nodes = yaml.load(node_yml, Loader=yaml.FullLoader)['nodes']

    if os.path.isfile('node_config.yml'):
        os.remove('node_config.yml')

    #should use ConfigLoader() here as well
    with open('node_config.yml', 'a') as node_configs:
        for node in nodes:
            node_type, node_name = node.split('.')
            if node_type == 'custom':
                node_config_path = os.path.join('src/custom_nodes', node_name, 'config.yml')
            else:
                dir_path = os.path.dirname(os.path.realpath(__file__))
                config_filename = node_name + '.yml'
                node_config_path = os.path.join(dir_path, 'configs', node_type, config_filename)
            if os.path.isfile(node_config_path):
                with open(node_config_path, 'r') as node_yml:
                    node_config = yaml.load(node_yml, Loader=yaml.FullLoader)
                    node_config = {node_name: node_config}
                yaml.dump(node_config, node_configs, default_flow_style=False)
            else:
                logger.info('No associated configs found for %s. Skipping.', node)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.input.utils.read import VideoThread


class Node(AbstractNode):
    """Node to receive livestream as inputs."""
    def __init__(self, config):
        super().__init__(config, node_path=__name__)

        resolution = config['resolution']
        input_source = config['input_source']
        mirror_image = config['mirror_image']

        self.videocap = VideoThread(resolution, input_source, mirror_image)

    def run(self, inputs: dict):
        success, img = self.videocap.read_frame()
        if success:
            outputs = {self.outputs[0]: img}
            return outputs

        raise Exception("An issue has been encountered reading the Image")



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os


def has_weights(root, path_to_check):
    """Checks for model weight paths from weights folder

    Args:
        root (str): path of peekingduck root folder
        path_to_check (List[str]): list of files/directories to check
            to see if weights exists

    Returns:
        boolean: True is files/directories needed exists, else False
    """

    # Check for whether weights dir even exist. If not make directory
    # Empty directory should then return False
    weights_dir = os.path.join(root, '..', 'weights')
    if not os.path.isdir(weights_dir):
        os.mkdir(weights_dir)
        return False

    for check in path_to_check:
        if not os.path.exists(os.path.join(root, check)):
            return False
    return True



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from .yolov4 import yolo_model


class Node(AbstractNode):
    """Yolo node class that initialises and use yolo model to infer bboxes
    from image frame
    """
    def __init__(self, config):
        super().__init__(config, node_path=__name__)
        self.model = yolo_model.YoloModel(config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that reads the image input and returns the bboxes
        of the specified objects chosen to be detected

        Args:
            inputs (Dict): Dictionary of inputs

        Returns:
            outputs (Dict): bbox output in dictionary format
        """
        # Currently prototyped to return just the bounding boxes
        # without the scores
        results, _, _ = self.model.predict(inputs[self.inputs[0]])
        outputs = {self.outputs[0]: results}
        return outputs



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
from typing import List, Dict, Any
import numpy as np
from peekingduck.weights_utils import checker, downloader
from .yolo_files.detector import Detector

class YoloModel:
    """Yolo model with model types: v3 and v3tiny"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check for yolo weights, if none then download into weights folder
        if not checker.has_weights(config['root'],
                                   config['weights_dir']):
            self.logger.info('---no yolo weights detected. proceeding to download...---')
            downloader.download_weights(config['root'],
                                        config['weights_id'])
            self.logger.info('---yolo weights download complete.---')

        #get classnames path to read all the classes
        classes_path = os.path.join(config['root'], config['classes'])
        self.class_names = [c.strip() for c in open(classes_path).readlines()]
        self.detect_ids = config['detect_ids']
        self.logger.info('yolo model detecting ids: %s', self.detect_ids)

        self.detector = Detector(config)

    def predict(self, frame: List[List[float]]) -> List[Any]:
        """predict the bbox from frame

        returns:
        object_bboxes(List[Numpy Array]): list of bboxes detected
        object_labels(List[str]): list of string labels of the
            object detected for the corresponding bbox
        object_scores(List(float)): list of confidence scores of the
            object detected for the corresponding bbox
        """
        assert isinstance(frame, np.ndarray)

        # return bboxes, object_bboxes, object_labels, object_scores
        return self.detector.predict_object_bbox_from_image(
            self.class_names, frame, self.detect_ids
        )

    def get_detect_ids(self) -> List[int]:
        """getter for selected ids for detection

        Returns:
            List[int]: list of selected detection ids
        """
        return self.detect_ids



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import tensorflow as tf


class BatchNormalization(tf.keras.layers.BatchNormalization):  # pylint: disable=too-many-ancestors, too-few-public-methods
    """
    Make trainable=False freeze BN for real (the og version is sad)
    """

    def call(self, inputs, training=None):
        """Make trainable=False freeze BN for real (the og version is sad)

        Input:
            - inputs: input image matrix
            - training: the training data

        Output:
            - inputs_image: transformed image matrix
        """
        if training is None:
            training = tf.constant(False)
        training = tf.logical_and(training, self.trainable)
        return super().call(inputs, training)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
import importlib
import logging
from typing import List, Dict, Any
import yaml
from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.config import ConfigLoader
from peekingduck.pipeline.nodes.node import AbstractNode

END_TYPE = 'process_end'

"""
Combine runner at this level. Use this to create the graph, and waddle is loop or once
"""

class Runner():
    """Runner class that uses the declared nodes to create pipeline to run inference
    """

    def __init__(self, RUN_PATH: str, CUSTOM_NODE_PATH: str = None,
                 nodes: List[AbstractNode] = None):
        """
        Args:
            RUN_PATH (str): path to yaml file of node pipeine declaration.
            CUSTOM_NODE_PATH (str): path to custom nodes folder if used.
            nodes (:obj:'list' of :obj:'Node'): if not using declarations via yaml,
                initialize by giving the node stack as a list

        """

        self.logger = logging.getLogger(__name__)

        instantiated_nodes = []

        if not nodes:
            with open(RUN_PATH) as file:
                self.run_config = yaml.load(file, Loader=yaml.FullLoader)
                self.logger.info(
                    'Successfully loaded run_config file. Proceeding to create Graph.')
            # create Graph to run
            nodes_config = ConfigLoader(self.run_config['nodes'])
            imported_nodes = []
            for node in self.run_config['nodes']:
                if node.split('.')[0] == 'custom':
                    custom_node_path = os.path.join(
                        CUSTOM_NODE_PATH, node.split('.')[1]+'.py')
                    spec = importlib.util.spec_from_file_location(
                        node, custom_node_path)
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                    imported_nodes.append(("custom", module))
                else:
                    imported_nodes.append((node, importlib.import_module(
                        'peekingduck.pipeline.nodes.' + node)))
                self.logger.info("%s added to pipeline.", node)

            # instantiate classes from imported nodes
            for node_name, node in imported_nodes:
                if node_name == 'custom':
                    instantiated_nodes.append(node.Node(None))
                else:
                    config = nodes_config.get(node_name)
                    instantiated_nodes.append(node.Node(config))

        # If Runner given nodes, instantiated_nodes is created differently
        else:
            for node in nodes:
                instantiated_nodes.append(node)

        # Create Graph
        try:
            self.pipeline = Pipeline(instantiated_nodes)
        except ValueError as exception:
            self.logger.error(str(exception))
            sys.exit(1)

    def run(self) -> None:
        """execute single or continuous inference
        """
        while not self.pipeline.video_end:
            self.pipeline.execute()

    def get_run_config(self) -> Dict[str, Any]:
        """retreive run configs

        Returns:
            Dict[Any]: run configs being used for runner
        """
        return self.run_config



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List
import tensorflow as tf


def transform_images(x_image: List, size: int) -> List:
    """transform image to size x size

    Input:
        - x_image: input image matrix
        - size: integer size of the image

    Output:
        - x_image: transformed image matrix
    """
    x_image = tf.image.resize(x_image, (size, size))
    x_image = x_image / 255
    return x_image



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.input.utils.read import VideoNoThread

class Node(AbstractNode):
    """Node to receive videos/image as inputs."""
    def __init__(self, config):
        super().__init__(config, node_path=__name__)
        self._allowed_extensions = ["jpg", "jpeg", "png", "mp4", "avi"]
        input_source = config['input_source']
        self._resolution = config['resolution']
        self._mirror_image = config['mirror_image']

        self._get_files(input_source)
        self._get_next_input()

    def run(self, inputs: dict):
        '''
        input: ["source"],
        output: ["img", "end"]
        '''
        outputs = self._run_single_file()

        if outputs["end"]:
            self._get_next_input()
            outputs = self._run_single_file()

        return outputs

    def _run_single_file(self) -> Dict[str, Any]:
        success, img = self.videocap.read_frame()

        outputs = {"img": None,
                   "end": True,
                   "filename": self._file_name,
                   "fps": self._fps}
        if success:
            outputs = {"img": img,
                       "end": False,
                       "filename": self._file_name,
                       "fps": self._fps}
        return outputs

    def _get_files(self, path) -> None:
        self._filepaths = [path]

        if os.path.isdir(path):
            self._filepaths = os.listdir(path)
            self._filepaths = [os.path.join(path, filepath)
                               for filepath in self._filepaths]
            self._filepaths.sort()

    def _get_next_input(self):

        if self._filepaths:
            file_path = self._filepaths.pop(0)
            self._file_name = os.path.basename(file_path)

            if self._is_valid_file_type(file_path):
                self.videocap = VideoNoThread(
                    self._resolution,
                    file_path,
                    self._mirror_image
                )
                self._fps = self.videocap.fps
            else:
                self.logger.warning("Skipping '%s' as it is not an accepted file format %s",
                                    file_path,
                                    str(self._allowed_extensions)
                                    )
                self._get_next_input()

    def _is_valid_file_type(self, filepath):

        if filepath.split(".")[-1] in self._allowed_extensions:
            return True
        return False



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
import cv2
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA


POSE_BBOX_COLOR = (255, 255, 0)
BLACK_COLOR = (0, 0, 0)
PINK_COLOR = (255, 0, 255)
ACTIVITY_COLOR = (100, 0, 255)
HUMAN_BBOX_COLOR = (255, 0, 255)
GROUP_BBOX_COLOR = (0, 140, 255)
HAND_BBOX_COLOR = (255, 0, 0)
OBJ_MASK_COLOR = (0, 100, 255)
KEYPOINT_TEXT_COLOR = (255, 0, 255)
KEYPOINT_DOT_COLOR = (0, 255, 0)
KEYPOINT_CONNECT_COLOR = (0, 255, 255)
HAND_KEYPOINT_DOT_COLOR = (0, 255, 0)
HAND_KEYPOINT_CONNECT_COLOR = (0, 0, 255)
COUNTING_TEXT_COLOR = (0, 0, 255)
FONT_SCALE = 1
FONT_THICKNESS = 2
SKELETON_SHORT_NAMES = (
    "N", "LEY", "REY", "LEA", "REA", "LSH",
    "RSH", "LEL", "REL", "LWR", "RWR",
    "LHI", "RHI", "LKN", "RKN", "LAN", "RAN")

def draw_human_poses(image, poses):
    '''draw pose estimates onto frame image'''
    image_size = _get_image_size(image)
    for pose in poses:
        if pose.bbox.shape == (2, 2):
            _draw_connections(image, pose.connections,
                              image_size, KEYPOINT_CONNECT_COLOR)
            _draw_keypoints(image, pose.keypoints,
                            pose.keypoint_scores, image_size,
                            KEYPOINT_DOT_COLOR)

def _get_image_size(frame):
    image_size = (frame.shape[1], frame.shape[0])  # width, height
    return image_size

def _draw_bbox(frame, bbox, image_size, color):
    top_left, bottom_right = _project_points_onto_original_image(bbox, image_size)
    cv2.rectangle(frame, (top_left[0], top_left[1]),
                  (bottom_right[0], bottom_right[1]),
                  color, 2)
    return top_left

def _draw_connections(frame, connections, image_size, connection_color):
    for connection in connections:
        pt1, pt2 = _project_points_onto_original_image(connection, image_size)
        cv2.line(frame, (pt1[0], pt1[1]), (pt2[0], pt2[1]), connection_color)

def _draw_keypoints(frame, keypoints, scores,
                    image_size, keypoint_dot_color):
    img_keypoints = _project_points_onto_original_image(
        keypoints, image_size)

    for idx, keypoint in enumerate(img_keypoints):
        _draw_one_keypoint_dot(frame, keypoint, keypoint_dot_color)
        if scores is not None:
            _draw_one_keypoint_text(frame, idx, keypoint)

def _draw_one_keypoint_dot(frame, keypoint, keypoint_dot_color):
    cv2.circle(frame, (keypoint[0], keypoint[1]), 5, keypoint_dot_color, -1)

def _draw_one_keypoint_text(frame, idx, keypoint):
    position = (keypoint[0], keypoint[1])
    text = str(SKELETON_SHORT_NAMES[idx])

    cv2.putText(frame, text, position, cv2.FONT_HERSHEY_SIMPLEX,
                0.4, KEYPOINT_TEXT_COLOR, 1, cv2.LINE_AA)

def _project_points_onto_original_image(points, image_size):
    """Project points from relative value to absolute values in original
    image.  E.g. from (1, 0.5) to (1280, 400).  It use a coordinate with
    original point (0, 0) at top-left.

    args:
        - points: np.array of (x, y) pairs of normalized joint coordinates.
                    i.e X and Y are in the range [0.0, 1.0]
        - image_size: image shape tuple to project (width, height)

    return:
        list of (x, y) pairs of joint coordinates transformed to image
        coordinates. x will be in the range [0, image width]. y will be in
        in the range [0, image height]
    """
    if len(points) == 0:
        return []

    points = points.reshape((-1, 2))

    projected_points = np.array(points, dtype=np.float32)
    width, height = image_size[0], image_size[1]
    projected_points[:, 0] *= width
    projected_points[:, 1] *= height

    return projected_points

def draw_human_bboxes(frame, human_bboxes):
    '''draw only human bboxes onto frame image'''
    image_size = _get_image_size(frame)
    for bbox in human_bboxes:
        _draw_bbox(frame, bbox, image_size, HUMAN_BBOX_COLOR)

def _draw_bbox(frame, bbox, image_size, color):
    top_left, bottom_right = _project_points_onto_original_image(bbox, image_size)
    cv2.rectangle(frame, (top_left[0], top_left[1]),
                  (bottom_right[0], bottom_right[1]),
                  color, 2)
    return top_left

def draw_fps(frame: np.array, current_fps: float) -> None:
    """ Draw FPS onto frame image

    args:
        - frame: array containing the RGB values of the frame image
        - current_fps: value of the calculated FPS
    """
    text = "FPS: {:.05}".format(current_fps)
    text_location = (25, 25)

    cv2.putText(frame, text, text_location, FONT_HERSHEY_SIMPLEX, FONT_SCALE,
                PINK_COLOR, FONT_THICKNESS, LINE_AA)



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import (
    Add,
    Concatenate,
    Conv2D,
    Input,
    Lambda,
    LeakyReLU,
    MaxPool2D,
    UpSampling2D,
    ZeroPadding2D,
)
from tensorflow.keras.regularizers import l2
from .batch_norm import BatchNormalization

# pylint: disable=redundant-keyword-arg, no-value-for-parameter, unexpected-keyword-arg, invalid-name, too-many-locals

YOLO_ANCHORS = np.array([(10, 13), (16, 30), (33, 23), (30, 61), (62, 45),
                         (59, 119), (116, 90), (156, 198),
                         (373, 326)], np.float32) / 416
YOLO_ANCHOR_MASKS = np.array([[6, 7, 8], [3, 4, 5], [0, 1, 2]])

YOLO_TINY_ANCHORS = np.array([(10, 14), (23, 27), (37, 58), (81, 82),
                              (135, 169), (344, 319)], np.float32) / 416
YOLO_TINY_ANCHOR_MASKS = np.array([[3, 4, 5], [0, 1, 2]])


def _darknet_conv(x, filters, size, strides=1, batch_norm=True):
    """create 1 layer with [padding], conv2d, [bn and relu]"""
    if strides == 1:
        padding = 'same'
    else:
        x = ZeroPadding2D(((1, 0), (1, 0)))(x)  # top left half-padding
        padding = 'valid'
    x = Conv2D(filters=filters,
               kernel_size=size,
               strides=strides,
               padding=padding,
               use_bias=not batch_norm,
               kernel_regularizer=l2(0.0005))(x)
    if batch_norm:
        x = BatchNormalization()(x)
        x = LeakyReLU(alpha=0.1)(x)
    return x


def _darknet_residual(x, filters):
    """create 2 layers by given H(x) = F(x) + x"""
    prev = x
    x = _darknet_conv(x, filters // 2, 1)
    x = _darknet_conv(x, filters, 3)
    x = Add()([prev, x])
    return x


def _darknet_block(x, filters, blocks):
    """create (1 + 2 x blocks) layers"""
    x = _darknet_conv(x, filters, 3, strides=2)
    for _ in range(blocks):
        x = _darknet_residual(x, filters)
    return x


def _darknet(name=None):
    """Create a Model with 52 layers

    This is different from classical Darknet, which shall be 53 layers,
    it can be calculated as below:
    1 + (1 + 2x1) + (1 + 2x2) + (1 + 2x8) + (1 + 2x8) + (1 + 2x4)

    Return:
        - tf.keras.Model with 3 outputs x_36, x_61 an x
    """
    x = inputs = Input([None, None, 3])
    x = _darknet_conv(x, 32, 3)
    x = _darknet_block(x, 64, 1)
    x = _darknet_block(x, 128, 2)  # skip connection
    x = x_36 = _darknet_block(x, 256, 8)  # skip connection
    x = x_61 = _darknet_block(x, 512, 8)
    x = _darknet_block(x, 1024, 4)
    return tf.keras.Model(inputs, (x_36, x_61, x), name=name)


def _darknet_tiny(name=None):
    """create 7 layers, return a Model with 2 outputs x_8 and x"""
    x = inputs = Input([None, None, 3])
    x = _darknet_conv(x, 16, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 32, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 64, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 128, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = x_8 = _darknet_conv(x, 256, 3)  # skip connection
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 512, 3)
    x = MaxPool2D(2, 1, 'same')(x)
    x = _darknet_conv(x, 1024, 3)
    return tf.keras.Model(inputs, (x_8, x), name=name)


def _yolo_conv(filters, name=None):
    """create convolution layers

    It has [1] + 5 layers

    Return:
        - Model with 1 output x
    """
    def _yolo_conv_imp(x_in):
        if isinstance(x_in, tuple):
            inputs = Input(x_in[0].shape[1:]), Input(x_in[1].shape[1:])
            x, x_skip = inputs

            # concat with skip connection
            x = _darknet_conv(x, filters, 1)
            x = UpSampling2D(2)(x)
            x = Concatenate()([x, x_skip])
        else:
            x = inputs = Input(x_in.shape[1:])

        x = _darknet_conv(x, filters, 1)
        x = _darknet_conv(x, filters * 2, 3)
        x = _darknet_conv(x, filters, 1)
        x = _darknet_conv(x, filters * 2, 3)
        x = _darknet_conv(x, filters, 1)
        return Model(inputs, x, name=name)(x_in)

    return _yolo_conv_imp


def _yolo_conv_tiny(filters, name=None):
    """create [1] layers, return a Model with 1 output x"""
    def _yolo_conv_tiny_imp(x_in):
        if isinstance(x_in, tuple):
            inputs = Input(x_in[0].shape[1:]), Input(x_in[1].shape[1:])
            x, x_skip = inputs

            # concat with skip connection
            x = _darknet_conv(x, filters, 1)
            x = UpSampling2D(2)(x)
            x = Concatenate()([x, x_skip])
        else:
            x = inputs = Input(x_in.shape[1:])
            x = _darknet_conv(x, filters, 1)

        return Model(inputs, x, name=name)(x_in)

    return _yolo_conv_tiny_imp


def _yolo_output(filters, anchors, classes, name=None):
    """create 2 layers, return a model with 1 output x"""
    def _yolo_output_imp(x_in):
        x = inputs = Input(x_in.shape[1:])
        x = _darknet_conv(x, filters * 2, 3)
        x = _darknet_conv(x, anchors * (classes + 5), 1, batch_norm=False)
        x = Lambda(lambda x: tf.reshape(x, (-1, tf.shape(x)[1], tf.shape(x)[2],
                                            anchors, classes + 5)))(x)
        return tf.keras.Model(inputs, x, name=name)(x_in)

    return _yolo_output_imp


def _yolo_boxes(pred, anchors, classes):
    # pred: (batch_size, grid, grid, anchors, (x, y, w, h, obj, ...classes))
    grid_size = tf.shape(pred)[1]
    box_xy, box_wh, objectness, class_probs = tf.split(pred,
                                                       (2, 2, 1, classes),
                                                       axis=-1)

    box_xy = tf.sigmoid(box_xy)
    objectness = tf.sigmoid(objectness)
    class_probs = tf.sigmoid(class_probs)
    pred_box = tf.concat((box_xy, box_wh), axis=-1)  # original xywh for loss

    # !!! grid[x][y] == (y, x)
    grid = tf.meshgrid(tf.range(grid_size), tf.range(grid_size))
    grid = tf.expand_dims(tf.stack(grid, axis=-1), axis=2)  # [gx, gy, 1, 2]

    box_xy = (box_xy + tf.cast(grid, tf.float32)) / \
        tf.cast(grid_size, tf.float32)
    box_wh = tf.exp(box_wh) * anchors

    box_x1y1 = box_xy - box_wh / 2
    box_x2y2 = box_xy + box_wh / 2
    bbox = tf.concat([box_x1y1, box_x2y2], axis=-1)

    return bbox, objectness, class_probs, pred_box


def _yolo_nms(outputs, classes):
    """non-maximum suppression"""
    # boxes, conf, type
    b, c, t = [], [], []

    for o in outputs:
        b.append(tf.reshape(o[0], (tf.shape(o[0])[0], -1, tf.shape(o[0])[-1])))
        c.append(tf.reshape(o[1], (tf.shape(o[1])[0], -1, tf.shape(o[1])[-1])))
        t.append(tf.reshape(o[2], (tf.shape(o[2])[0], -1, tf.shape(o[2])[-1])))

    bbox = tf.concat(b, axis=1)
    confidence = tf.concat(c, axis=1)
    class_probs = tf.concat(t, axis=1)

    scores = confidence * class_probs
    boxes, scores, classes, valid_detections = tf.image.combined_non_max_suppression(
        boxes=tf.reshape(bbox, (tf.shape(bbox)[0], -1, 1, 4)),
        scores=tf.reshape(scores,
                          (tf.shape(scores)[0], -1, tf.shape(scores)[-1])),
        max_output_size_per_class=100,
        max_total_size=100,
        iou_threshold=0.5,
        score_threshold=0.2)

    return boxes, scores, classes, valid_detections


def yolov3(size=None,
           channels=3,
           classes=80,
           training=False):
    """Create a yolov3 model

    This model has 76 layers, which can be calcualted below:
    1 darknet (52 layers) +
    3 yolo_conv (6 layers) +
    3 yolo_output (2 layers)

    Args:
        - size:     (int) input image size
        - channels: (int) channel for input image, for RBG it is 3
        - anchors:  (np.array) the boxes used to detect object, with a series of (x, y)
                    values, in this case it is 9 normalized (x, y) values.
        - masks:    (np.array) masks for anchors to decide which anchors will be used,
                    in this case it is 3 of (i, k, j) values.
        - classes:  (int) the number of classes of objects this Yolov3 can detect
        - training: (boolean) whether it is in training or not

    Returns:
        - (tensorflow.keras.Model) with inputs and outputs (3 items in outputs)
    """
    anchors = YOLO_ANCHORS
    masks = YOLO_ANCHOR_MASKS

    x = inputs = Input([size, size, channels])

    x_36, x_61, x = _darknet(name='yolo_darknet')(x)

    x = _yolo_conv(512, name='yolo_conv_0')(x)
    output_0 = _yolo_output(512, len(masks[0]), classes, name='yolo_output_0')(x)

    x = _yolo_conv(256, name='yolo_conv_1')((x, x_61))
    output_1 = _yolo_output(256, len(masks[1]), classes, name='yolo_output_1')(x)

    x = _yolo_conv(128, name='yolo_conv_2')((x, x_36))
    output_2 = _yolo_output(128, len(masks[2]), classes, name='yolo_output_2')(x)

    if training:
        return Model(inputs, (output_0, output_1, output_2), name='yolov3')

    boxes_0 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[0]], classes),
                     name='yolo_boxes_0')(output_0)
    boxes_1 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[1]], classes),
                     name='yolo_boxes_1')(output_1)
    boxes_2 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[2]], classes),
                     name='yolo_boxes_2')(output_2)

    outputs = Lambda(lambda x: _yolo_nms(x, classes),
                     name='yolo_nms')((boxes_0[:3], boxes_1[:3], boxes_2[:3]))

    return Model(inputs, outputs, name='yolov3')


def yolov3_tiny(size=None,
                channels=3,
                classes=80,
                training=False):
    """Create Yolov3 tiny model

    This model has 23 layers, which can be calcualted below:
    1 darknet_tiny (7 layers) +
    2 yolo_conv (6 layers) +
    2 yolo_output (2 layers)

    Input:
        - size:     (int) input image size
        - channels: (int) channel for input image, for RBG it is 3
        - anchors:  (np.array) the boxes used to detect object, with a series of (x, y)
                    values, in this case it is 9 of (x, y) values
        - masks:    (np.array) masks for anchors to decide which anchors will be used,
                    in this case it is 3 of (i, k, j) values
        - classes:  (int) the number of classes of objects this Yolov3 can detect
        - training: (boolean) whether it is in training or not

    Return:
        - (tensorflow.keras.Model) with inputs and outputs (2 items in outputs)
    """
    anchors = YOLO_TINY_ANCHORS
    masks = YOLO_TINY_ANCHOR_MASKS

    x = inputs = Input([size, size, channels])

    x_8, x = _darknet_tiny(name='yolo_darknet')(x)

    x = _yolo_conv_tiny(256, name='yolo_conv_0')(x)
    output_0 = _yolo_output(256, len(masks[0]), classes,
                            name='yolo_output_0')(x)

    x = _yolo_conv_tiny(128, name='yolo_conv_1')((x, x_8))
    output_1 = _yolo_output(128, len(masks[1]), classes,
                            name='yolo_output_1')(x)

    if training:
        return Model(inputs, (output_0, output_1), name='yolov3')

    boxes_0 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[0]], classes),
                     name='yolo_boxes_0')(output_0)
    boxes_1 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[1]], classes),
                     name='yolo_boxes_1')(output_1)
    outputs = Lambda(lambda x: _yolo_nms(x, classes),
                     name='yolo_nms')((boxes_0[:3], boxes_1[:3]))

    return Model(inputs, outputs, name='yolov3_tiny')



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
from typing import List
import tensorflow as tf

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
SAVE_DIR = os.path.join(os.getcwd(), 'data', 'yolov3')

logger = logging.getLogger(__name__) #pylint: disable=invalid-name

def wrap_frozen_graph(graph_def: tf.compat.v1.GraphDef, inputs: List, outputs: List):
    '''
    Wraps the graph into a function. This is akin to a model.predict() function
    in keras. When doing inference, simply do frozen_function(tf.cast(x, float))[0].
    It will return your predicted values.

    args:
        - graph_def: The frozen graph in graph_def format
        - inputs: The name(s) of the input nodes from your graph. e.g.['inputs']
        - outputs: The name(s) of your output nodes from your graph.
                    e.g. ['heatmap', 'offsets', 'displacement_fwd', 'displacement_bwd']
        - print_graph: Whether to print the graph

    return:
        a wrapped_import function to perform your inference with.
    '''
    def _imports_graph_def():  # this needs to be here because of graph_def
        tf.compat.v1.import_graph_def(graph_def, name="")

    wrapped_import = tf.compat.v1.wrap_function(_imports_graph_def, [])
    import_graph = wrapped_import.graph

    return wrapped_import.prune(
        tf.nest.map_structure(import_graph.as_graph_element, inputs),
        tf.nest.map_structure(import_graph.as_graph_element, outputs))


def load_graph(filename: str, inputs: List, outputs: List):
    '''
    Loads the graph
    '''
    with tf.io.gfile.GFile(filename, "rb") as graph_file:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(graph_file.read())

        # print_inputs(graph_def)
        # print_outputs(graph_def)

        frozen_func = wrap_frozen_graph(graph_def=graph_def,
                                        inputs=inputs,
                                        outputs=outputs)

        return frozen_func


def print_inputs(graph_def: tf.compat.v1.GraphDef):
    '''
    Prints the input nodes of graph_def
    '''
    # pylint: disable=not-context-manager
    with tf.Graph().as_default() as graph:
        tf.import_graph_def(graph_def, name='')

    input_list = []
    for operation in graph.get_operations():  # tensorflow.python.framework.ops.Operation
        if operation.type == "Placeholder":
            input_list.append(operation.name)

        logger.info('Inputs: %s', input_list)


def print_outputs(graph_def: tf.compat.v1.GraphDef):
    '''
    Prints the output nodes of graph_def
    '''
    name_list = []
    input_list = []
    for node in graph_def.node:  # tensorflow.core.framework.node_def_pb2.NodeDef
        name_list.append(node.name)
        input_list.extend(node.input)

    outputs = set(name_list) - set(input_list)
    logger.info('Outputs: %s', list(outputs))



"""
Copyright 2021 AI Singapore

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This is detection file that will be used to detect person bounding box with
the confident score.

Public function:
    - detect_person(): return bounding box, score for person
    - detect_person_and_plot_to_output_image(): plot output image

Commands:
    python src/yolov3/detector.py
    python src/yolov3/detector.py --tiny
"""
import os
import logging
import numpy as np
import tensorflow as tf
from .graph_functions import load_graph
from .dataset import transform_images


class Detector:
    """Object detection class using yolo model to find object bboxes"""

    def __init__(self, config):
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.root_dir = config['root']

        self.yolo = self._create_yolo_model()

    def _setup_gpu(self):
        physical_devices = tf.config.experimental.list_physical_devices('GPU')
        if len(physical_devices) > 0:
            self.logger.info('GPU setup with %d devices', len(physical_devices))
        else:
            self.logger.info('use CPU')

    def _create_yolo_model(self):
        '''
        Creates yolo model for human detection
        '''
        model_type = self.config['model_type']
        if self.config['yolo_graph_mode']:
            model_path = os.path.join(self.root_dir, self.config['graph_files'][model_type])
            return self._load_yolo_graph(model_path)
        model_path = os.path.join(self.root_dir, self.config['model_files'][model_type])
        model = tf.keras.models.load_model(model_path)
        return model

    def _load_yolo_graph(self, filepath):
        '''
        When loading a graph model, you need to explicitly state the input
        and output nodes of the graph. It is usually x:0 for input and Identity:0
        for outputs, depending on how many output nodes you have.
        '''
        model_type = 'yolo%s' % self.config['model_type'][:2]
        model_nodes = self.config['MODEL_NODES'][model_type]
        model_path = os.path.join(filepath)
        if os.path.isfile(model_path):
            return load_graph(model_path, inputs=model_nodes['inputs'],
                              outputs=model_nodes['outputs'])
        raise ValueError('Graph file does not exist. Please check that '
                         '%s exists' % model_path)

    def _load_image(self, image_file):
        img = open(image_file, 'rb').read()
        self.logger.info('image file %s loaded', image_file)
        return img

    @staticmethod
    def _reshape_image(image, image_size):
        image = tf.expand_dims(image, 0)
        image = transform_images(image, image_size)
        return image

    @staticmethod
    def _shrink_dimension_and_length(boxes, scores, classes, nums, object_ids):
        len0 = nums[0]

        classes = classes.numpy()[0]
        classes = classes[:len0]
        mask1 = np.isin(classes, tuple(object_ids))  # only identify objects we are interested in
        classes = tf.boolean_mask(classes, mask1)

        scores = scores.numpy()[0]
        scores = scores[:len0]
        scores = tf.boolean_mask(scores, mask1)

        boxes = boxes.numpy()[0]
        boxes = boxes[:len0]
        boxes = tf.boolean_mask(boxes, mask1)

        return boxes, scores, classes

    def _evaluate_image_by_yolo(self, image):
        '''
        Takes in the yolo model and image to perform inference with.
        It will return the following:
            - boxes: the bounding boxes for each object
            - scores: the scores for each object predicted
            - classes: the class predicted for each bounding box
            - nums: number of valid bboxes. Only nums[0] should be used. The rest
                    are paddings.
        '''
        # image = image[..., ::-1]  # swap from bgr to rgb
        pred = self.yolo(image)[-1]
        bboxes = pred[:, :, :4].numpy()
        bboxes[:, :, [0, 1]] = bboxes[:, :, [1, 0]]  # swapping x and y axes
        bboxes[:, :, [2, 3]] = bboxes[:, :, [3, 2]]
        pred_conf = pred[:, :, 4:]

        # performs nms using model's predictions
        boxes, scores, classes, nums = tf.image.combined_non_max_suppression(
            boxes=tf.reshape(bboxes, (tf.shape(bboxes)[0], -1, 1, 4)),
            scores=tf.reshape(
                pred_conf, (tf.shape(pred_conf)[0], -1, tf.shape(pred_conf)[-1])),
            max_output_size_per_class=50,
            max_total_size=50,
            iou_threshold=self.config['yolo_iou_threshold'],
            score_threshold=self.config['yolo_score_threshold']
        )
        return boxes, scores, classes, nums

    @staticmethod
    def _prepare_image_from_camera(image):
        image = image.astype(np.float32)
        image = tf.convert_to_tensor(image)
        return image

    @staticmethod
    def _prepare_image_from_file(image):
        image = tf.image.decode_image(image, channels=3)
        return image

    def predict_person_bbox_from_image(self, image):
        """Detect all persons' bounding box from one image

        args:
            - yolo:  (Model) model like yolov3 or yolov3_tiny
            - image: (np.array) input image

        return:
            - boxes: (np.array) an array of bounding box with
                    definition like (x1, y1, x2, y2), in a
                    coordinate system with original point in
                    the left top corner
        """
        # 1. prepare image
        image = self._prepare_image_from_camera(image)
        image = self._reshape_image(image, image_size=self.config['size'])

        # 2. evaluate image
        boxes, scores, classes, nums = self._evaluate_image_by_yolo(image)

        # 3. clean up return. Note that object id is [0] as index 0 is for person
        boxes, scores, classes = self._shrink_dimension_and_length(
            boxes, scores, classes, nums, [0])

        boxes = np.array(boxes)
        return boxes

    # possible that we may want to control what is being detection
    def predict_object_bbox_from_image(self, class_names, image, detect_ids):
        """Detect all objects' bounding box from one image

        args:
            - yolo:  (Model) model like yolov3 or yolov3_tiny
            - image: (np.array) input image

        return:
            - boxes: (np.array) an array of bounding box with
                    definition like (x1, y1, x2, y2), in a
                    coordinate system with original point in
                    the left top corner
        """
        # 1. prepare image
        image = self._prepare_image_from_camera(image)
        image = self._reshape_image(image, self.config['size'])

        # 2. evaluate image
        boxes, scores, classes, nums = self._evaluate_image_by_yolo(image)

        # 3. clean up return
        boxes, scores, classes = self._shrink_dimension_and_length(
            boxes, scores, classes, nums, detect_ids)

        # convert classes into class names
        classes = [class_names[int(i)] for i in classes]

        boxes = np.array(boxes)
        return boxes, classes, scores
